//---------------------------------------------------------------------------
#include <vcl.h>
#pragma hdrstop

#include "FarPlugin.h"
#include <Common.h>
//---------------------------------------------------------------------------
TCustomFarPlugin * FarPlugin = NULL;
#define FAR_TITLE_SUFFIX " - Far"
//---------------------------------------------------------------------------
#pragma package(smart_init)
//---------------------------------------------------------------------------
__fastcall TCustomFarPlugin::TCustomFarPlugin(THandle AHandle): TObject()
{
  FHandle = AHandle;

  FProgressSaveScreenHandle = 0;

  FOpenedPlugins = new TObjectList();
  FOpenedPlugins->OwnsObjects = false;
  FSavedTitles = new TStringList();

  memset(&FPluginInfo, 0, sizeof(FPluginInfo));
  ClearPluginInfo(FPluginInfo);

  // far\Examples\Compare\compare.cpp
  FConsoleInput = CreateFile("CONIN$", GENERIC_READ, FILE_SHARE_READ, NULL,
    OPEN_EXISTING, 0, NULL);
  FConsoleOutput = CreateFile("CONOUT$", GENERIC_READ | GENERIC_WRITE,
    FILE_SHARE_READ | FILE_SHARE_WRITE, NULL, OPEN_EXISTING, 0, NULL);
}
//---------------------------------------------------------------------------
__fastcall TCustomFarPlugin::~TCustomFarPlugin()
{
  CloseHandle(FConsoleInput);
  FConsoleInput = INVALID_HANDLE_VALUE;
  CloseHandle(FConsoleOutput);
  FConsoleOutput = INVALID_HANDLE_VALUE;

  if (FProgressSaveScreenHandle)
  {
    ClearProgress();
  }
  ClearPluginInfo(FPluginInfo);
  assert(FOpenedPlugins->Count == 0);
  delete FOpenedPlugins;
  delete FSavedTitles;
}
//---------------------------------------------------------------------------
bool __fastcall TCustomFarPlugin::HandlesFunction(THandlesFunction /*Function*/)
{
  return false;
}
//---------------------------------------------------------------------------
int __fastcall TCustomFarPlugin::GetMinFarVersion()
{
  return 0;
}
//---------------------------------------------------------------------------
void __fastcall TCustomFarPlugin::SetStartupInfo(const struct PluginStartupInfo * Info)
{
  try
  {
    assert(Info->StructSize >= sizeof(FStartupInfo));
    memcpy(&FStartupInfo, Info, sizeof(FStartupInfo));
    assert(Info->FSF->StructSize >= sizeof(FFarStandardFunctions));
    memcpy(&FFarStandardFunctions, Info->FSF, sizeof(FFarStandardFunctions));
  }
  catch(Exception & E)
  {
    HandleException(&E);
  }
}
//---------------------------------------------------------------------------
void __fastcall TCustomFarPlugin::ExitFAR()
{
}
//---------------------------------------------------------------------------
void __fastcall TCustomFarPlugin::GetPluginInfo(struct PluginInfo * Info)
{
  try
  {
    Info->StructSize = sizeof(PluginInfo);
    TStrings * DiskMenuStrings = NULL;
    TStrings * PluginMenuStrings = NULL;
    TStrings * PluginConfigStrings = NULL;
    TStrings * CommandPrefixes = NULL;
    try
    {
      DiskMenuStrings = new TStringList();
      PluginMenuStrings = new TStringList();
      PluginConfigStrings = new TStringList();
      CommandPrefixes = new TStringList();

      ClearPluginInfo(FPluginInfo);

      GetPluginInfoEx(FPluginInfo.Flags, DiskMenuStrings, PluginMenuStrings,
        PluginConfigStrings, CommandPrefixes);

      #define COMPOSESTRINGARRAY(NAME) \
        if (NAME->Count) \
        { \
          char ** StringArray = new char *[NAME->Count]; \
          FPluginInfo.NAME = StringArray; \
          FPluginInfo.NAME ## Number = NAME->Count; \
          for (int Index = 0; Index < NAME->Count; Index++) \
          { \
            StringArray[Index] = DuplicateStr(NAME->Strings[Index]); \
          } \
        }

      COMPOSESTRINGARRAY(DiskMenuStrings);
      COMPOSESTRINGARRAY(PluginMenuStrings);
      COMPOSESTRINGARRAY(PluginConfigStrings);

      #undef COMPOSESTRINGARRAY(NAME)

      if (DiskMenuStrings->Count)
      {
        int * NumberArray = new int[DiskMenuStrings->Count];
        FPluginInfo.DiskMenuNumbers = NumberArray;
        for (int Index = 0; Index < DiskMenuStrings->Count; Index++)
        {
          NumberArray[Index] = (int)DiskMenuStrings->Objects[Index];
        }
      }

      AnsiString CommandPrefix;
      for (int Index = 0; Index < CommandPrefixes->Count; Index++)
      {
        CommandPrefix = CommandPrefix + (CommandPrefix.IsEmpty() ? "" : ":") +
          CommandPrefixes->Strings[Index];
      }
      FPluginInfo.CommandPrefix = DuplicateStr(CommandPrefix);
    }
    __finally
    {
      delete DiskMenuStrings;
      delete PluginMenuStrings;
      delete PluginConfigStrings;
      delete CommandPrefixes;
    }
    memcpy(Info, &FPluginInfo, sizeof(FPluginInfo));
  }
  catch(Exception & E)
  {
    HandleException(&E);
  }
}
//---------------------------------------------------------------------------
AnsiString __fastcall TCustomFarPlugin::GetModuleName()
{
  return FStartupInfo.ModuleName;
}
//---------------------------------------------------------------------------
void __fastcall TCustomFarPlugin::ClearPluginInfo(PluginInfo & Info)
{
  if (Info.StructSize)
  {
    #define FREESTRINGARRAY(NAME) \
      for (int Index = 0; Index < Info.NAME ## Number; Index++) \
      { \
        delete Info.NAME[Index]; \
      } \
      delete[] Info.NAME;

    FREESTRINGARRAY(DiskMenuStrings);
    FREESTRINGARRAY(PluginMenuStrings);
    FREESTRINGARRAY(PluginConfigStrings);

    #undef FREESTRINGARRAY

    delete Info.CommandPrefix;
  }
  memset(&Info, 0, sizeof(Info));
  Info.StructSize = sizeof(Info);
}
//---------------------------------------------------------------------------
char * TCustomFarPlugin::DuplicateStr(const AnsiString Str, bool AllowEmpty)
{
  if (Str.IsEmpty() && !AllowEmpty)
  {
    return NULL;
  }
  else
  {
    char * Result = new char[Str.Length() + 1];
    strcpy(Result, Str.c_str());
    return Result;
  }
}
//---------------------------------------------------------------------------
TCustomFarFileSystem * __fastcall TCustomFarPlugin::GetPanelFileSystem(bool Another)
{
  TCustomFarFileSystem * Result = NULL;
  PanelInfo Info;
  FarControl(Another ? FCTL_GETANOTHERPANELINFO : FCTL_GETPANELINFO, &Info);

  if (Info.Plugin)
  {
    TRect Bounds = Info.PanelRect;
    TCustomFarFileSystem * FileSystem;
    int Index = 0;
    while (!Result && Index < FOpenedPlugins->Count)
    {
      FileSystem = dynamic_cast<TCustomFarFileSystem *>(FOpenedPlugins->Items[Index]);
      assert(FileSystem);
      if (FileSystem->PanelInfo->Bounds == Bounds)
      {
        Result = FileSystem;
      }
      Index++;
    }
  }

  return Result;
}
//---------------------------------------------------------------------------
int __fastcall TCustomFarPlugin::Configure(int Item)
{
  try
  {
    int Result = ConfigureEx(Item);

    return Result;
  }
  catch(Exception & E)
  {
    HandleException(&E);
    return false;
  }
}
//---------------------------------------------------------------------------
void * __fastcall TCustomFarPlugin::OpenPlugin(int OpenFrom, int Item)
{
  try
  {
    TCustomFarFileSystem * Result = OpenPluginEx(OpenFrom, Item);

    if (Result)
    {
      FOpenedPlugins->Add(Result);
    }
    else
    {
      Result = (TCustomFarFileSystem *)INVALID_HANDLE_VALUE;
    }

    return Result;
  }
  catch(Exception & E)
  {
    HandleException(&E);
    return INVALID_HANDLE_VALUE;
  }
}
//---------------------------------------------------------------------------
void __fastcall TCustomFarPlugin::ClosePlugin(void * Plugin)
{
  try
  {
    assert(FOpenedPlugins->IndexOf(static_cast<TObject*>(Plugin)) >= 0);
    try
    {
      static_cast<TCustomFarFileSystem*>(Plugin)->FClosed = true;
      delete static_cast<TObject*>(Plugin);
    }
    __finally
    {
      FOpenedPlugins->Remove(static_cast<TObject*>(Plugin));
    }
  }
  catch(Exception & E)
  {
    HandleException(&E);
  }
}
//---------------------------------------------------------------------------
void __fastcall TCustomFarPlugin::GetOpenPluginInfo(HANDLE Plugin,
  struct OpenPluginInfo * Info)
{
  try
  {
    assert(FOpenedPlugins->IndexOf((TObject*)Plugin) >= 0);
    ((TCustomFarFileSystem *)Plugin)->GetOpenPluginInfo(Info);
  }
  catch(Exception & E)
  {
    ((TCustomFarFileSystem *)Plugin)->HandleException(&E);
  }
}
//---------------------------------------------------------------------------
int __fastcall TCustomFarPlugin::GetFindData(HANDLE Plugin,
  struct PluginPanelItem ** PanelItem, int * ItemsNumber, int OpMode)
{
  try
  {
    assert(FOpenedPlugins->IndexOf((TObject*)Plugin) >= 0);
    return ((TCustomFarFileSystem *)Plugin)->GetFindData(PanelItem, ItemsNumber, OpMode);
  }
  catch(Exception & E)
  {
    ((TCustomFarFileSystem *)Plugin)->HandleException(&E, OpMode);
    return 0;
  }
}
//---------------------------------------------------------------------------
void __fastcall TCustomFarPlugin::FreeFindData(HANDLE Plugin,
  struct PluginPanelItem * PanelItem, int ItemsNumber)
{
  try
  {
    assert(FOpenedPlugins->IndexOf((TObject*)Plugin) >= 0);
    ((TCustomFarFileSystem *)Plugin)->FreeFindData(PanelItem, ItemsNumber);
  }
  catch(Exception & E)
  {
    ((TCustomFarFileSystem *)Plugin)->HandleException(&E);
  }
}
//---------------------------------------------------------------------------
int __fastcall TCustomFarPlugin::ProcessHostFile(HANDLE Plugin,
  struct PluginPanelItem * PanelItem, int ItemsNumber, int OpMode)
{
  try
  {
    if (HandlesFunction(hfProcessHostFile))
    {
      assert(FOpenedPlugins->IndexOf((TObject*)Plugin) >= 0);
      return ((TCustomFarFileSystem *)Plugin)->ProcessHostFile(
        PanelItem, ItemsNumber, OpMode);
    }
    else
    {
      return 0;
    }
  }
  catch(Exception & E)
  {
    ((TCustomFarFileSystem *)Plugin)->HandleException(&E, OpMode);
    return 0;
  }
}
//---------------------------------------------------------------------------
int __fastcall TCustomFarPlugin::ProcessKey(HANDLE Plugin, int Key,
  unsigned int ControlState)
{
  try
  {
    if (HandlesFunction(hfProcessKey))
    {
      assert(FOpenedPlugins->IndexOf((TObject*)Plugin) >= 0);
      return ((TCustomFarFileSystem *)Plugin)->ProcessKey(Key, ControlState);
    }
    else
    {
      return 0;
    }
  }
  catch(Exception & E)
  {
    ((TCustomFarFileSystem *)Plugin)->HandleException(&E);
    // when error occused, assume that key can be handled by plugin and
    // should not be processed by FAR
    return 1;
  }
}
//---------------------------------------------------------------------------
int __fastcall TCustomFarPlugin::ProcessEvent(HANDLE Plugin, int Event, void * Param)
{
  try
  {
    if (HandlesFunction(hfProcessEvent))
    {
      assert(FOpenedPlugins->IndexOf((TObject*)Plugin) >= 0);
      return ((TCustomFarFileSystem *)Plugin)->ProcessEvent(Event, Param);
    }
    else
    {
      return false;
    }
  }
  catch(Exception & E)
  {
    ((TCustomFarFileSystem *)Plugin)->HandleException(&E);
    return Event == FE_COMMAND ? true : false;
  }
}
//---------------------------------------------------------------------------
int __fastcall TCustomFarPlugin::SetDirectory(HANDLE Plugin, const char * Dir, int OpMode)
{
  try
  {
    assert(FOpenedPlugins->IndexOf((TObject*)Plugin) >= 0);
    return ((TCustomFarFileSystem *)Plugin)->SetDirectory(Dir, OpMode);
  }
  catch(Exception & E)
  {
    ((TCustomFarFileSystem *)Plugin)->HandleException(&E, OpMode);
    return 0;
  }
}
//---------------------------------------------------------------------------
int __fastcall TCustomFarPlugin::MakeDirectory(HANDLE Plugin, char * Name, int OpMode)
{
  try
  {
    assert(FOpenedPlugins->IndexOf((TObject*)Plugin) >= 0);
    return ((TCustomFarFileSystem *)Plugin)->MakeDirectory(Name, OpMode);
  }
  catch(Exception & E)
  {
    ((TCustomFarFileSystem *)Plugin)->HandleException(&E, OpMode);
    return 0;
  }
}
//---------------------------------------------------------------------------
int __fastcall TCustomFarPlugin::DeleteFiles(HANDLE Plugin,
  struct PluginPanelItem *PanelItem, int ItemsNumber, int OpMode)
{
  try
  {
    assert(FOpenedPlugins->IndexOf((TObject*)Plugin) >= 0);
    return ((TCustomFarFileSystem *)Plugin)->DeleteFiles(PanelItem, ItemsNumber, OpMode);
  }
  catch(Exception & E)
  {
    ((TCustomFarFileSystem *)Plugin)->HandleException(&E, OpMode);
    return 0;
  }
}
//---------------------------------------------------------------------------
int __fastcall TCustomFarPlugin::GetFiles(HANDLE Plugin,
  struct PluginPanelItem *PanelItem, int ItemsNumber, int Move,
  char * DestPath, int OpMode)
{
  try
  {
    assert(FOpenedPlugins->IndexOf((TObject*)Plugin) >= 0);
    return ((TCustomFarFileSystem *)Plugin)->GetFiles(PanelItem, ItemsNumber,
      Move, DestPath, OpMode);
  }
  catch(Exception & E)
  {
    // display error even for OPM_FIND
    ((TCustomFarFileSystem *)Plugin)->HandleException(&E, OpMode & ~OPM_FIND);
    return 0;
  }
}
//---------------------------------------------------------------------------
int __fastcall TCustomFarPlugin::PutFiles(HANDLE Plugin,
  struct PluginPanelItem *PanelItem, int ItemsNumber, int Move, int OpMode)
{
  try
  {
    assert(FOpenedPlugins->IndexOf((TObject*)Plugin) >= 0);
    return ((TCustomFarFileSystem *)Plugin)->PutFiles(PanelItem, ItemsNumber,
      Move, OpMode);
  }
  catch(Exception & E)
  {
    ((TCustomFarFileSystem *)Plugin)->HandleException(&E, OpMode);
    return 0;
  }
}
//---------------------------------------------------------------------------
int __fastcall TCustomFarPlugin::Message(unsigned int Flags,
  const AnsiString Title, const AnsiString Message, TStrings * Buttons)
{
  int TabSize = 8;
  int Result;
  if (Buttons)
  {
    TStrings * MessageLines = NULL;
    TStrings * Lines = NULL;
    TStrings * WrappedLines = NULL;
    char ** Items = NULL;
    try
    {
      Lines = new TStringList();
      Lines->Text = Message;
      WrappedLines = new TStringList();
      MessageLines = new TStringList();
      MessageLines->Add(Title);
      for (int Index = 0; Index < Lines->Count; Index++)
      {
        AnsiString WrappedLine = Lines->Strings[Index];
        if (!WrappedLine.IsEmpty())
        {
          WrappedLine = StringReplace(WrappedLine, "'", "\3", TReplaceFlags() << rfReplaceAll);
          WrappedLine = WrapText(WrappedLine, MaxMessageWidth);
          WrappedLine = StringReplace(WrappedLine, "\3", "'", TReplaceFlags() << rfReplaceAll);
          WrappedLines->Text = WrappedLine;
          for (int WrappedIndex = 0; WrappedIndex < WrappedLines->Count; WrappedIndex++)
          {
            AnsiString Line = WrappedLines->Strings[WrappedIndex];
            int P;
            while ((P = Line.Pos("\t")) > 0)
            {
              Line.Delete(P, 1);
              Line.Insert(AnsiString::StringOfChar(' ',
                  ((P / TabSize) + ((P % TabSize) > 0 ? 1 : 0)) * TabSize - P + 1),
                P);
            }
            MessageLines->Add(Line);
          }
        }
        else
        {
          MessageLines->Add("");
        }
      }
      for (int Index = 0; Index < Buttons->Count; Index++)
      {
        MessageLines->Add(Buttons->Strings[Index]);
      }

      Items = new char *[MessageLines->Count];
      for (int Index = 0; Index < MessageLines->Count; Index++)
      {
        Items[Index] = MessageLines->Strings[Index].c_str();
      }

      Result = FStartupInfo.Message(FStartupInfo.ModuleNumber,
        Flags | FMSG_LEFTALIGN, NULL, Items, MessageLines->Count,
        Buttons->Count);
    }
    __finally
    {
      delete Items;
      delete Lines;
      delete MessageLines;
      delete WrappedLines;
    }
  }
  else
  {
    AnsiString Items = Title + "\n" + Message;
    Result = FStartupInfo.Message(FStartupInfo.ModuleNumber,
      Flags | FMSG_ALLINONE | FMSG_LEFTALIGN, NULL, (char **)Items.c_str(), 0, 0);
  }
  return Result;
}
//---------------------------------------------------------------------------
int __fastcall TCustomFarPlugin::Menu(unsigned int Flags, const AnsiString Title,
  const AnsiString Bottom, const FarMenuItem * Items, int Count,
  const int * BreakKeys, int & BreakCode)
{
  assert(Items);
  return FStartupInfo.Menu(FStartupInfo.ModuleNumber, -1, -1, 0,
    Flags, (char*)Title.data(), (char*)Bottom.data(), NULL, BreakKeys,
    &BreakCode, Items, Count);
}
//---------------------------------------------------------------------------
int __fastcall TCustomFarPlugin::Menu(unsigned int Flags, const AnsiString Title,
  const AnsiString Bottom, TStrings * Items, const int * BreakKeys,
  int & BreakCode)
{
  assert(Items && Items->Count);
  int Result;
  FarMenuItemEx * MenuItems = new FarMenuItemEx[Items->Count];
  try
  {
    int Selected = -1;
    for (int i = 0; i < Items->Count; i++)
    {
      memset(&MenuItems[i], 0, sizeof(MenuItems[i]));
      AnsiString Text = Items->Strings[i].SubString(1, sizeof(MenuItems[i].Text)-1);
      MenuItems[i].Flags = int(Items->Objects[i]);
      if (MenuItems[i].Flags & MIF_SELECTED)
      {
        assert(Selected < 0);
        Selected = i;
      }
      strcpy(MenuItems[i].Text.Text, Text.c_str());
    }

    Result = Menu(Flags | FMENU_USEEXT, Title, Bottom,
      (const FarMenuItem *)MenuItems, Items->Count, BreakKeys, BreakCode);

    if (Result >= 0)
    {
      if (Selected >= 0)
      {
        Items->Objects[Selected] = (TObject *)(int(Items->Objects[Selected]) & !MIF_SELECTED);
      }
      Items->Objects[Result] = (TObject *)(int(Items->Objects[Result]) | MIF_SELECTED);
    }
  }
  __finally
  {
    delete MenuItems;
  }
  return Result;
}
//---------------------------------------------------------------------------
int __fastcall TCustomFarPlugin::Menu(unsigned int Flags, const AnsiString Title,
  const AnsiString Bottom, TStrings * Items)
{
  int BreakCode;
  return Menu(Flags, Title, Bottom, Items, NULL, BreakCode);
}
//---------------------------------------------------------------------------
bool __fastcall TCustomFarPlugin::InputBox(const AnsiString Title,
  const AnsiString Prompt, AnsiString & Text, unsigned long Flags,
  const AnsiString HistoryName, int MaxLen)
{
  int Result;
  AnsiString DestText;
  DestText.SetLength(MaxLen + 1);
  Result = FStartupInfo.InputBox(Title.c_str(), Prompt.c_str(),
    (char*)HistoryName.data(), Text.c_str(), DestText.c_str(), MaxLen, NULL,
    FIB_ENABLEEMPTY | FIB_BUTTONS | Flags);
  if (Result)
  {
    Text = DestText.c_str();
  }
  return (Result != 0);
}
//---------------------------------------------------------------------------
void __fastcall TCustomFarPlugin::ShowProgress(const AnsiString Message)
{
  if (FProgressSaveScreenHandle)
  {
    RestoreScreen(FProgressSaveScreenHandle);
    UpdateConsoleTitle(Message);
  }
  else
  {
    ShowConsoleTitle(Message);
  }
  SaveScreen(FProgressSaveScreenHandle);
  AnsiString Items = AnsiString("\n") + Message;
  FStartupInfo.Message(FStartupInfo.ModuleNumber,
    FMSG_ALLINONE, NULL, (char**)Items.c_str(), 0, 0);
}
//---------------------------------------------------------------------------
void __fastcall TCustomFarPlugin::ClearProgress()
{
  ClearConsoleTitle();
  RestoreScreen(FProgressSaveScreenHandle);
}
//---------------------------------------------------------------------------
void __fastcall TCustomFarPlugin::Text(int X, int Y, int Color, const AnsiString Str)
{
  FStartupInfo.Text(X, Y, Color, Str.c_str());
}
//---------------------------------------------------------------------------
void __fastcall TCustomFarPlugin::FlushText()
{
  FStartupInfo.Text(0, 0, 0, NULL);
}
//---------------------------------------------------------------------------
void __fastcall TCustomFarPlugin::WriteConsole(const AnsiString Str)
{
  unsigned long Written;
  ::WriteConsole(FConsoleOutput, Str.c_str(), Str.Length(), &Written, NULL);
}
//---------------------------------------------------------------------------
void __fastcall TCustomFarPlugin::FarCopyToClipboard(const AnsiString Str)
{
  FFarStandardFunctions.CopyToClipboard(Str.c_str());
}
//---------------------------------------------------------------------------
void __fastcall TCustomFarPlugin::ScrollTerminalScreen(int Rows)
{
  CONSOLE_SCREEN_BUFFER_INFO BufferInfo;
  GetConsoleScreenBufferInfo(FConsoleOutput, &BufferInfo);

  SMALL_RECT Source;
  COORD Dest;
  CHAR_INFO Fill;
  Source.Left = 0;
  Source.Top = (char)Rows;
  Source.Right = BufferInfo.dwSize.X;
  Source.Bottom = BufferInfo.dwSize.Y;
  Dest.X = 0;
  Dest.Y = 0;
  Fill.Char.AsciiChar = ' ';
  Fill.Attributes = 7;
  ScrollConsoleScreenBuffer(FConsoleOutput, &Source, NULL, Dest, &Fill);
}
//---------------------------------------------------------------------------
void __fastcall TCustomFarPlugin::ShowTerminalScreen()
{
  CONSOLE_SCREEN_BUFFER_INFO BufferInfo;
  GetConsoleScreenBufferInfo(FConsoleOutput, &BufferInfo);

  AnsiString Blank = AnsiString::StringOfChar(' ', BufferInfo.dwSize.X);
  for (int Y = 0; Y< BufferInfo.dwSize.Y; Y++)
  {
    Text(0, Y, 7/* LIGHTGRAY */, Blank);
  }
  FlushText();

  COORD Coord;
  Coord.X = 0;
  Coord.Y = BufferInfo.dwCursorPosition.Y;
  SetConsoleCursorPosition(FConsoleOutput, Coord);
}
//---------------------------------------------------------------------------
void __fastcall TCustomFarPlugin::ShowConsoleTitle(const AnsiString Title)
{
  char SaveTitle[512];
  GetConsoleTitle(SaveTitle, sizeof(SaveTitle));
  AnsiString SaveTitleStripped = SaveTitle;
  if (SaveTitleStripped.SubString(SaveTitleStripped.Length() -
        AnsiString(FAR_TITLE_SUFFIX).Length() + 1, 6) == FAR_TITLE_SUFFIX)
  {
    SaveTitleStripped.SetLength(SaveTitleStripped.Length() -
      AnsiString(FAR_TITLE_SUFFIX).Length());
  }
  FSavedTitles->AddObject(SaveTitle,
    (TObject *)(!FCurrentTitle.IsEmpty() && FCurrentTitle == SaveTitleStripped));
  FCurrentTitle = Title;
  FCurrentProgress = -1;
  UpdateConsoleTitle();
}
//---------------------------------------------------------------------------
void __fastcall TCustomFarPlugin::ClearConsoleTitle()
{
  assert(FSavedTitles->Count > 0);
  FCurrentTitle = "";
  FCurrentProgress = -1;
  AnsiString Title = FSavedTitles->Strings[FSavedTitles->Count-1];
  SetConsoleTitle(Title.c_str());
  if (FSavedTitles->Objects[FSavedTitles->Count-1])
  {
    FCurrentTitle = Title;
  }
  FSavedTitles->Delete(FSavedTitles->Count-1);
}
//---------------------------------------------------------------------------
void __fastcall TCustomFarPlugin::UpdateConsoleTitle(const AnsiString Title)
{
  assert(!FCurrentTitle.IsEmpty());
  FCurrentTitle = Title;
  UpdateConsoleTitle();
}
//---------------------------------------------------------------------------
void __fastcall TCustomFarPlugin::UpdateConsoleTitleProgress(int Progress)
{
  assert(!FCurrentTitle.IsEmpty());
  FCurrentProgress = Progress;
  UpdateConsoleTitle();
}
//---------------------------------------------------------------------------
void __fastcall TCustomFarPlugin::UpdateConsoleTitle()
{
  AnsiString Title;
  if (FCurrentProgress >= 0)
  {
    Title = FORMAT("{%d%%} %s", (FCurrentProgress, FCurrentTitle));
  }
  else
  {
    Title = FCurrentTitle;
  }
  Title += FAR_TITLE_SUFFIX;
  SetConsoleTitle(Title.c_str());
}
//---------------------------------------------------------------------------
void __fastcall TCustomFarPlugin::SaveScreen(THandle & Screen)
{
  assert(!Screen);
  Screen = (THandle)FStartupInfo.SaveScreen(0, 0, -1, -1);
  assert(Screen);
}
//---------------------------------------------------------------------------
void __fastcall TCustomFarPlugin::RestoreScreen(THandle & Screen)
{
  assert(Screen);
  FStartupInfo.RestoreScreen((HANDLE)Screen);
  Screen = 0;
}
//---------------------------------------------------------------------------
void __fastcall TCustomFarPlugin::HandleException(Exception * E, int /*OpMode*/)
{
  assert(E);
  Message(FMSG_WARNING | FMSG_MB_OK, "", E->Message);
}
//---------------------------------------------------------------------------
AnsiString __fastcall TCustomFarPlugin::GetMsg(int MsgId)
{
  return FStartupInfo.GetMsg(FStartupInfo.ModuleNumber, MsgId);
}
//---------------------------------------------------------------------------
bool __fastcall TCustomFarPlugin::CheckForEsc()
{
  INPUT_RECORD Rec;
  unsigned long ReadCount;
  while (PeekConsoleInput(FConsoleInput, &Rec, 1, &ReadCount) && ReadCount)
  {
    ReadConsoleInput(FConsoleInput, &Rec, 1, &ReadCount);
    if (Rec.EventType == KEY_EVENT &&
        Rec.Event.KeyEvent.wVirtualKeyCode == VK_ESCAPE &&
        Rec.Event.KeyEvent.bKeyDown)
    {
      return true;
    }
  }
  return false;
}
//---------------------------------------------------------------------------
bool __fastcall TCustomFarPlugin::Viewer(const AnsiString FileName,
  unsigned int Flags, const AnsiString Title)
{
  int Result = FStartupInfo.Viewer(FileName.c_str(),
    static_cast<const char*>(Title.data()), 0, 0, -1, -1, Flags);
  return Result;
}
//---------------------------------------------------------------------------
void __fastcall TCustomFarPlugin::FarControl(int Command, void * Param)
{
  int Result = FStartupInfo.Control(INVALID_HANDLE_VALUE, Command, Param);
  assert(Result);
}
//---------------------------------------------------------------------------
int __fastcall TCustomFarPlugin::FarAdvControl(int Command, void * Param)
{
  return FStartupInfo.AdvControl(FStartupInfo.ModuleNumber, Command, Param);
}
//---------------------------------------------------------------------------
//---------------------------------------------------------------------------
__fastcall TCustomFarFileSystem::TCustomFarFileSystem(TCustomFarPlugin * APlugin):
    TObject()
{
  FPlugin = APlugin;
  FPanelInfo = NULL;
  FClosed = false;

  memset(&FOpenPluginInfo, 0, sizeof(FOpenPluginInfo));
  ClearOpenPluginInfo(FOpenPluginInfo);
};
//---------------------------------------------------------------------------
__fastcall TCustomFarFileSystem::~TCustomFarFileSystem()
{
  ResetPanelInfo();
  ClearOpenPluginInfo(FOpenPluginInfo);
}
//---------------------------------------------------------------------------
void __fastcall TCustomFarFileSystem::HandleException(Exception * E, int OpMode)
{
  FPlugin->HandleException(E, OpMode);
}
//---------------------------------------------------------------------------
void __fastcall TCustomFarFileSystem::ClearOpenPluginInfo(OpenPluginInfo & Info)
{
  if (Info.StructSize)
  {
    delete Info.HostFile;
    delete Info.CurDir;
    delete Info.Format;
    delete Info.PanelTitle;
    assert(!Info.InfoLines);
    assert(!Info.InfoLinesNumber);
    assert(!Info.DescrFiles);
    assert(!Info.DescrFilesNumber);
    assert(Info.PanelModesNumber == 0 || Info.PanelModesNumber == PANEL_MODES_COUNT);
    for (int Index = 0; Index < Info.PanelModesNumber; Index++)
    {
      assert(Info.PanelModesArray);
      TFarPanelModes::ClearPanelMode(Info.PanelModesArray[Index]);
    }
    delete Info.PanelModesArray;
    if (Info.KeyBar)
    {
      TFarKeyBarTitles::ClearKeyBarTitles(*Info.KeyBar);
      delete Info.KeyBar;
    }
    delete Info.ShortcutData;
  }
  memset(&Info, 0, sizeof(Info));
  Info.StructSize = sizeof(Info);
}
//---------------------------------------------------------------------------
void __fastcall TCustomFarFileSystem::GetOpenPluginInfo(struct OpenPluginInfo * Info)
{
  ResetPanelInfo();
  ClearOpenPluginInfo(FOpenPluginInfo);
  AnsiString HostFile, CurDir, Format, PanelTitle, ShortcutData;
  bool StartSortOrder;
  TFarPanelModes * PanelModes = NULL;
  TFarKeyBarTitles * KeyBarTitles = NULL;
  try
  {
    PanelModes = new TFarPanelModes();
    KeyBarTitles = new TFarKeyBarTitles();
    StartSortOrder = false;

    GetOpenPluginInfoEx(FOpenPluginInfo.Flags, HostFile, CurDir, Format,
      PanelTitle, PanelModes, FOpenPluginInfo.StartPanelMode,
      FOpenPluginInfo.StartSortMode, StartSortOrder, KeyBarTitles, ShortcutData);

    FOpenPluginInfo.HostFile = TCustomFarPlugin::DuplicateStr(HostFile);
    FOpenPluginInfo.CurDir = TCustomFarPlugin::DuplicateStr(CurDir);
    FOpenPluginInfo.Format = TCustomFarPlugin::DuplicateStr(Format);
    FOpenPluginInfo.PanelTitle = TCustomFarPlugin::DuplicateStr(PanelTitle);
    PanelModes->FillOpenPluginInfo(&FOpenPluginInfo);
    FOpenPluginInfo.StartSortOrder = StartSortOrder;
    KeyBarTitles->FillOpenPluginInfo(&FOpenPluginInfo);
    FOpenPluginInfo.ShortcutData = TCustomFarPlugin::DuplicateStr(ShortcutData);
    
    memcpy(Info, &FOpenPluginInfo, sizeof(FOpenPluginInfo));
  }
  __finally
  {
    delete PanelModes;
    delete KeyBarTitles;
  }
}
//---------------------------------------------------------------------------
int __fastcall TCustomFarFileSystem::GetFindData(
  struct PluginPanelItem ** PanelItem, int * ItemsNumber, int /*OpMode*/)
{
  ResetPanelInfo();
  TObjectList * PanelItems = new TObjectList();
  bool Result;
  try
  {
    Result = !FClosed && GetFindDataEx(PanelItems);
    if (Result && PanelItems->Count)
    {
      *PanelItem = new PluginPanelItem[PanelItems->Count];
      memset(*PanelItem, 0, PanelItems->Count * sizeof(PluginPanelItem));
      *ItemsNumber = PanelItems->Count;
      for (int Index = 0; Index < PanelItems->Count; Index++)
      {
        ((TCustomFarPanelItem *)PanelItems->Items[Index])->FillPanelItem(
          &((*PanelItem)[Index]));
      }
    }
    else
    {
      *PanelItem = NULL;
      *ItemsNumber = 0;
    }
  }
  __finally
  {
    delete PanelItems;
  }
  return Result;
}
//---------------------------------------------------------------------------
void __fastcall TCustomFarFileSystem::FreeFindData(
  struct PluginPanelItem * PanelItem, int ItemsNumber)
{
  ResetPanelInfo();
  if (PanelItem)
  {
    assert(ItemsNumber > 0);
    for (int Index = 0; Index < ItemsNumber; Index++)
    {
      delete PanelItem[Index].Description;
      delete PanelItem[Index].Owner;
      for (int CustomIndex = 0; CustomIndex < PanelItem[Index].CustomColumnNumber; CustomIndex++)
      {
        delete PanelItem[Index].CustomColumnData[CustomIndex];
      }
      delete PanelItem[Index].CustomColumnData;
    }
    delete PanelItem;
  }
}
//---------------------------------------------------------------------------
int __fastcall TCustomFarFileSystem::ProcessHostFile(struct PluginPanelItem * PanelItem,
  int ItemsNumber, int OpMode)
{
  ResetPanelInfo();
  TList * PanelItems = CreatePanelItemList(PanelItem, ItemsNumber);
  bool Result;
  try
  {
    Result = ProcessHostFileEx(PanelItems, OpMode);
  }
  __finally
  {
    delete PanelItems;
  }

  return Result;
}
//---------------------------------------------------------------------------
int __fastcall TCustomFarFileSystem::ProcessKey(int Key, unsigned int ControlState)
{
  ResetPanelInfo();
  return ProcessKeyEx(Key, ControlState);
}
//---------------------------------------------------------------------------
int __fastcall TCustomFarFileSystem::ProcessEvent(int Event, void * Param)
{
  ResetPanelInfo();
  return ProcessEventEx(Event, Param);
}
//---------------------------------------------------------------------------
int __fastcall TCustomFarFileSystem::SetDirectory(const char * Dir, int OpMode)
{
  ResetPanelInfo();
  return SetDirectoryEx(Dir, OpMode);
}
//---------------------------------------------------------------------------
int __fastcall TCustomFarFileSystem::MakeDirectory(char * Name, int OpMode)
{
  ResetPanelInfo();
  AnsiString NameStr = Name;
  int Result;
  try
  {
    Result = MakeDirectoryEx(NameStr, OpMode);
  }
  __finally
  {
    if (NameStr != Name)
    {
      strcpy(Name, NameStr.c_str());
    }
  }
  return Result;
}
//---------------------------------------------------------------------------
int __fastcall TCustomFarFileSystem::DeleteFiles(struct PluginPanelItem * PanelItem,
    int ItemsNumber, int OpMode)
{
  ResetPanelInfo();
  TList * PanelItems = CreatePanelItemList(PanelItem, ItemsNumber);
  bool Result;
  try
  {
    Result = DeleteFilesEx(PanelItems, OpMode);
  }
  __finally
  {
    delete PanelItems;
  }

  return Result;
}
//---------------------------------------------------------------------------
int __fastcall TCustomFarFileSystem::GetFiles(struct PluginPanelItem * PanelItem,
    int ItemsNumber, int Move, char * DestPath, int OpMode)
{
  ResetPanelInfo();
  TList * PanelItems = CreatePanelItemList(PanelItem, ItemsNumber);
  int Result;
  AnsiString DestPathStr = DestPath;
  try
  {
    Result = GetFilesEx(PanelItems, Move, DestPathStr, OpMode);
  }
  __finally
  {
    if (DestPathStr != DestPath)
    {
      strcpy(DestPath, DestPathStr.c_str());
    }
    delete PanelItems;
  }

  return Result;
}
//---------------------------------------------------------------------------
int __fastcall TCustomFarFileSystem::PutFiles(struct PluginPanelItem *PanelItem,
  int ItemsNumber, int Move, int OpMode)
{
  ResetPanelInfo();
  TList * PanelItems = CreatePanelItemList(PanelItem, ItemsNumber);
  int Result;
  try
  {
    Result = PutFilesEx(PanelItems, Move, OpMode);
  }
  __finally
  {
    delete PanelItems;
  }

  return Result;
}
//---------------------------------------------------------------------------
void __fastcall TCustomFarFileSystem::ResetPanelInfo()
{
  if (FPanelInfo)
  {
    SAFE_DESTROY(FPanelInfo);
  }
}
//---------------------------------------------------------------------------
TFarPanelInfo * __fastcall TCustomFarFileSystem::GetPanelInfo()
{
  if (!FPanelInfo)
  {
    ::PanelInfo * Info = new ::PanelInfo;
    FarControl(FCTL_GETPANELINFO, Info);
    FPanelInfo = new TFarPanelInfo(Info, this);
  }
  return FPanelInfo;
}
//---------------------------------------------------------------------------
void __fastcall TCustomFarFileSystem::FarControl(int Command, void * Param)
{
  int Result = FPlugin->FStartupInfo.Control(this, Command, Param);
  assert(Result);
}
//---------------------------------------------------------------------------
void __fastcall TCustomFarFileSystem::UpdatePanel(bool ClearSelection, bool Another)
{
  FarControl(Another ? FCTL_UPDATEANOTHERPANEL : FCTL_UPDATEPANEL,
    (void *)(!ClearSelection));
}
//---------------------------------------------------------------------------
void __fastcall TCustomFarFileSystem::RedrawPanel(bool Another)
{
  FarControl(Another ? FCTL_REDRAWANOTHERPANEL : FCTL_REDRAWPANEL, NULL);
}
//---------------------------------------------------------------------------
void __fastcall TCustomFarFileSystem::ClosePlugin()
{
  FClosed = true;
  FarControl(FCTL_CLOSEPLUGIN, NULL);
//  UpdatePanel();
}
//---------------------------------------------------------------------------
AnsiString __fastcall TCustomFarFileSystem::GetMsg(int MsgId)
{
  return FPlugin->GetMsg(MsgId);
}
//---------------------------------------------------------------------------
bool __fastcall TCustomFarFileSystem::ProcessHostFileEx(TList * /*PanelItems*/, int /*OpMode*/)
{
  return false;
}
//---------------------------------------------------------------------------
bool __fastcall TCustomFarFileSystem::ProcessKeyEx(int /*Key*/, unsigned int /*ControlState*/)
{
  return false;
}
//---------------------------------------------------------------------------
bool __fastcall TCustomFarFileSystem::ProcessEventEx(int /*Event*/, void * /*Param*/)
{
  return false;
}
//---------------------------------------------------------------------------
bool __fastcall TCustomFarFileSystem::SetDirectoryEx(const AnsiString /*Dir*/, int /*OpMode*/)
{
  return false;
}
//---------------------------------------------------------------------------
int __fastcall TCustomFarFileSystem::MakeDirectoryEx(AnsiString & /*Name*/, int /*OpMode*/)
{
  return -1;
}
//---------------------------------------------------------------------------
bool __fastcall TCustomFarFileSystem::DeleteFilesEx(TList * /*PanelItems*/, int /*OpMode*/)
{
  return false;
}
//---------------------------------------------------------------------------
int __fastcall TCustomFarFileSystem::GetFilesEx(TList * /*PanelItems*/, bool /*Move*/,
  AnsiString & /*DestPath*/, int /*OpMode*/)
{
  return 0;
}
//---------------------------------------------------------------------------
int __fastcall TCustomFarFileSystem::PutFilesEx(TList * /*PanelItems*/,
  bool /*Move*/, int /*OpMode*/)
{
  return 0;
}
//---------------------------------------------------------------------------
TList * __fastcall TCustomFarFileSystem::CreatePanelItemList(
  struct PluginPanelItem * PanelItem, int ItemsNumber)
{
  TList * PanelItems = new TObjectList();
  try
  {
    for (int Index = 0; Index < ItemsNumber; Index++)
    {
      PanelItems->Add(new TFarPanelItem(&PanelItem[Index]));
    }
  }
  catch(...)
  {
    delete PanelItems;
    throw;
  }
  return PanelItems;
}
//---------------------------------------------------------------------------
__fastcall TFarPanelModes::TFarPanelModes() : TObject()
{
  memset(FPanelModes, 0, sizeof(FPanelModes));
  FReferenced = false;
}
//---------------------------------------------------------------------------
__fastcall TFarPanelModes::~TFarPanelModes()
{
  if (!FReferenced)
  {
    for (int Index = 0; Index < LENOF(FPanelModes); Index++)
    {
      ClearPanelMode(FPanelModes[Index]);
    }
  }
}
//---------------------------------------------------------------------------
void __fastcall TFarPanelModes::SetPanelMode(int Mode, const AnsiString ColumnTypes,
  const AnsiString ColumnWidths, TStrings * ColumnTitles,
  bool FullScreen, bool DetailedStatus, bool AlignExtensions,
  bool CaseConversion, const AnsiString StatusColumnTypes,
  const AnsiString StatusColumnWidths)
{
  int ColumnTypesCount = !ColumnTypes.IsEmpty() ? CommaCount(ColumnTypes)+1 : 0;
  assert(Mode >= 0 && Mode < LENOF(FPanelModes));
  //assert(CommaCount(ColumnTypes) == CommaCount(ColumnWidths));
  assert(!ColumnTitles || (ColumnTitles->Count == ColumnTypesCount));
  //assert(CommaCount(StatusColumnTypes) == CommaCount(StatusColumnWidths));

  ClearPanelMode(FPanelModes[Mode]);
  FPanelModes[Mode].ColumnTypes = TCustomFarPlugin::DuplicateStr(ColumnTypes);
  FPanelModes[Mode].ColumnWidths = TCustomFarPlugin::DuplicateStr(ColumnWidths);
  if (ColumnTitles)
  {
    FPanelModes[Mode].ColumnTitles = new char *[ColumnTypesCount];
    for (int Index = 0; Index < ColumnTypesCount; Index++)
    {
      FPanelModes[Mode].ColumnTitles[Index] = TCustomFarPlugin::DuplicateStr(
        ColumnTitles->Strings[Index]);
    }
  }
  FPanelModes[Mode].FullScreen = FullScreen;
  FPanelModes[Mode].DetailedStatus = DetailedStatus;
  FPanelModes[Mode].AlignExtensions = AlignExtensions;
  FPanelModes[Mode].CaseConversion = CaseConversion;

  FPanelModes[Mode].StatusColumnTypes = TCustomFarPlugin::DuplicateStr(StatusColumnTypes);
  FPanelModes[Mode].StatusColumnWidths = TCustomFarPlugin::DuplicateStr(StatusColumnWidths);
}
//---------------------------------------------------------------------------
void __fastcall TFarPanelModes::ClearPanelMode(PanelMode & Mode)
{
  if (Mode.ColumnTypes)
  {
    int ColumnTypesCount = Mode.ColumnTypes ?
      CommaCount(AnsiString(Mode.ColumnTypes)) + 1 : 0;

    delete Mode.ColumnTypes;
    delete Mode.ColumnWidths;
    if (Mode.ColumnTitles)
    {
      for (int Index = 0; Index < ColumnTypesCount; Index++)
      {
        delete Mode.ColumnTitles[Index];
      }
      delete Mode.ColumnTitles;
    }
    delete Mode.StatusColumnTypes;
    delete Mode.StatusColumnWidths;
    memset(&Mode, 0, sizeof(Mode));
  }
}
//---------------------------------------------------------------------------
void __fastcall TFarPanelModes::FillOpenPluginInfo(struct OpenPluginInfo * Info)
{
  assert(Info);
  Info->PanelModesNumber = LENOF(FPanelModes);
  Info->PanelModesArray = new PanelMode[LENOF(FPanelModes)];
  memcpy(Info->PanelModesArray, &FPanelModes, sizeof(FPanelModes));
  FReferenced = true;
}
//---------------------------------------------------------------------------
int __fastcall TFarPanelModes::CommaCount(const AnsiString ColumnTypes)
{
  int Count = 0;
  for (int Index = 1; Index <= ColumnTypes.Length(); Index++)
  {
    if (ColumnTypes[Index] == ',')
    {
      Count++;
    }
  }
  return Count;
}
//---------------------------------------------------------------------------
__fastcall TFarKeyBarTitles::TFarKeyBarTitles()
{
  memset(&FKeyBarTitles, 0, sizeof(FKeyBarTitles));
  FReferenced = false;
}
//---------------------------------------------------------------------------
__fastcall TFarKeyBarTitles::~TFarKeyBarTitles()
{
  if (!FReferenced)
  {
    ClearKeyBarTitles(FKeyBarTitles);
  }
}
//---------------------------------------------------------------------------
void __fastcall TFarKeyBarTitles::ClearFileKeyBarTitles()
{
  ClearKeyBarTitle(fsNone, 3, 8);
  ClearKeyBarTitle(fsCtrl, 4, 11);
  ClearKeyBarTitle(fsAlt, 3, 7);
  ClearKeyBarTitle(fsShift, 1, 8);
  ClearKeyBarTitle(fsCtrlShift, 3, 4);
}
//---------------------------------------------------------------------------
void __fastcall TFarKeyBarTitles::ClearKeyBarTitle(TFarShiftStatus ShiftStatus,
  int FunctionKeyStart, int FunctionKeyEnd)
{
  if (!FunctionKeyEnd)
  {
    FunctionKeyEnd = FunctionKeyStart;
  }
  for (int Index = FunctionKeyStart; Index <= FunctionKeyEnd; Index++)
  {
    SetKeyBarTitle(ShiftStatus, Index, "");
  }
}
//---------------------------------------------------------------------------
void __fastcall TFarKeyBarTitles::SetKeyBarTitle(TFarShiftStatus ShiftStatus,
  int FunctionKey, const AnsiString Title)
{
  assert(FunctionKey >= 1 && FunctionKey <= LENOF(FKeyBarTitles.Titles));
  char ** Titles;
  switch (ShiftStatus) {
    case fsNone: Titles = FKeyBarTitles.Titles; break;
    case fsCtrl: Titles = FKeyBarTitles.CtrlTitles; break;
    case fsAlt: Titles = FKeyBarTitles.AltTitles; break;
    case fsShift: Titles = FKeyBarTitles.ShiftTitles; break;
    case fsCtrlShift: Titles = FKeyBarTitles.CtrlShiftTitles; break;
    case fsAltShift: Titles = FKeyBarTitles.AltShiftTitles; break;
    case fsCtrlAlt: Titles = FKeyBarTitles.CtrlAltTitles; break;
    default: assert(false);
  }
  if (Titles[FunctionKey-1])
  {
    delete Titles[FunctionKey-1];
  }
  Titles[FunctionKey-1] = TCustomFarPlugin::DuplicateStr(Title, true);
}
//---------------------------------------------------------------------------
void __fastcall TFarKeyBarTitles::ClearKeyBarTitles(KeyBarTitles & Titles)
{
  for (int Index = 0; Index < LENOF(Titles.Titles); Index++)
  {
    delete Titles.Titles[Index];
    delete Titles.CtrlTitles[Index];
    delete Titles.AltTitles[Index];
    delete Titles.ShiftTitles[Index];
    delete Titles.CtrlShiftTitles[Index];
    delete Titles.AltShiftTitles[Index];
    delete Titles.CtrlAltTitles[Index];
  }
}
//---------------------------------------------------------------------------
void __fastcall TFarKeyBarTitles::FillOpenPluginInfo(struct OpenPluginInfo * Info)
{
  assert(Info);
  Info->KeyBar = new KeyBarTitles;
  memcpy(Info->KeyBar, &FKeyBarTitles, sizeof(FKeyBarTitles));
  FReferenced = true;
}
//---------------------------------------------------------------------------
//---------------------------------------------------------------------------
AnsiString __fastcall TCustomFarPanelItem::CustomColumnData(int /*Column*/)
{
  assert(false);
  return "";
}
//---------------------------------------------------------------------------
void __fastcall TCustomFarPanelItem::FillPanelItem(struct PluginPanelItem * PanelItem)
{
  assert(PanelItem);

  AnsiString FileName;
  __int64 Size = 0;
  TDateTime LastWriteTime;
  TDateTime LastAccess;
  AnsiString Description;
  AnsiString Owner;

  GetData(PanelItem->Flags, FileName, Size, PanelItem->FindData.dwFileAttributes,
    LastWriteTime, LastAccess, PanelItem->NumberOfLinks, Description, Owner,
    (void *)PanelItem->UserData, PanelItem->CustomColumnNumber);

  FILETIME FileTime = DateTimeToFileTime(LastWriteTime);
  FILETIME FileTimeA = DateTimeToFileTime(LastAccess);
  PanelItem->FindData.ftCreationTime = FileTime;
  PanelItem->FindData.ftLastAccessTime = FileTimeA;
  PanelItem->FindData.ftLastWriteTime = FileTime;
  PanelItem->FindData.nFileSizeLow = (long int)Size;
  PanelItem->FindData.nFileSizeHigh = (long int)(Size >> 32);
  PanelItem->PackSize = (long int)Size;
  ASCOPY(PanelItem->FindData.cFileName, FileName);
  PanelItem->Description = TCustomFarPlugin::DuplicateStr(Description);
  PanelItem->Owner = TCustomFarPlugin::DuplicateStr(Owner);

  PanelItem->CustomColumnData = new char *[PanelItem->CustomColumnNumber];
  for (int Index = 0; Index < PanelItem->CustomColumnNumber; Index++)
  {
    PanelItem->CustomColumnData[Index] =
      TCustomFarPlugin::DuplicateStr(CustomColumnData(Index));
  }
}
//---------------------------------------------------------------------------
//---------------------------------------------------------------------------
__fastcall TFarPanelItem::TFarPanelItem(PluginPanelItem * APanelItem):
  TCustomFarPanelItem()
{
  assert(APanelItem);
  FPanelItem = APanelItem;
}
//---------------------------------------------------------------------------
void __fastcall TFarPanelItem::GetData(
  unsigned long & /*Flags*/, AnsiString & /*FileName*/, __int64 & /*Size*/,
  unsigned long & /*FileAttributes*/,
  TDateTime & /*LastWriteTime*/, TDateTime & /*LastAccess*/,
  unsigned long & /*NumberOfLinks*/, AnsiString & /*Description*/,
  AnsiString & /*Owner*/, void *& /*UserData*/, int & /*CustomColumnNumber*/)
{
  assert(false);
}
//---------------------------------------------------------------------------
AnsiString __fastcall TFarPanelItem::CustomColumnData(int /*Column*/)
{
  assert(false);
  return "";
}
//---------------------------------------------------------------------------
unsigned long __fastcall TFarPanelItem::GetFlags()
{
  return FPanelItem->Flags;
}
//---------------------------------------------------------------------------
AnsiString __fastcall TFarPanelItem::GetFileName()
{
  return FPanelItem->FindData.cFileName;
}
//---------------------------------------------------------------------------
void * __fastcall TFarPanelItem::GetUserData()
{
  return (void *)FPanelItem->UserData;
}
//---------------------------------------------------------------------------
bool __fastcall TFarPanelItem::GetSelected()
{
  return (FPanelItem->Flags & PPIF_SELECTED) != 0;
}
//---------------------------------------------------------------------------
void __fastcall TFarPanelItem::SetSelected(bool value)
{
  if (value)
  {
    FPanelItem->Flags |= PPIF_SELECTED;
  }
  else
  {
    FPanelItem->Flags &= ~PPIF_SELECTED;
  }
}
//---------------------------------------------------------------------------
unsigned long __fastcall TFarPanelItem::GetFileAttributes()
{
  return FPanelItem->FindData.dwFileAttributes;
}
//---------------------------------------------------------------------------
bool __fastcall TFarPanelItem::GetIsParentDirectory()
{
  return (FileName == "..");
}
//---------------------------------------------------------------------------
bool __fastcall TFarPanelItem::GetIsFile()
{
  return (FileAttributes & FILE_ATTRIBUTE_DIRECTORY) == 0;
}
//---------------------------------------------------------------------------
//---------------------------------------------------------------------------
__fastcall THintPanelItem::THintPanelItem(const AnsiString AHint) :
  TCustomFarPanelItem()
{
  FHint = AHint;
}
//---------------------------------------------------------------------------
void __fastcall THintPanelItem::GetData(
  unsigned long & /*Flags*/, AnsiString & FileName, __int64 & /*Size*/,
  unsigned long & /*FileAttributes*/,
  TDateTime & /*LastWriteTime*/, TDateTime & /*LastAccess*/,
  unsigned long & /*NumberOfLinks*/, AnsiString & /*Description*/,
  AnsiString & /*Owner*/, void *& /*UserData*/, int & /*CustomColumnNumber*/)
{
  FileName = FHint;
}
//---------------------------------------------------------------------------
//---------------------------------------------------------------------------
__fastcall TFarPanelInfo::TFarPanelInfo(PanelInfo * APanelInfo, TCustomFarFileSystem * AOwner):
  TObject()
{
  assert(APanelInfo);
  FPanelInfo = APanelInfo;
  FOwner = AOwner;
  FItems = NULL;
}
//---------------------------------------------------------------------------
__fastcall TFarPanelInfo::~TFarPanelInfo()
{
  delete FPanelInfo;
  delete FItems;
}
//---------------------------------------------------------------------------
int __fastcall TFarPanelInfo::GetItemCount()
{
  return FPanelInfo->ItemsNumber;
}
//---------------------------------------------------------------------------
TRect __fastcall TFarPanelInfo::GetBounds()
{
  return FPanelInfo->PanelRect;
}
//---------------------------------------------------------------------------
int __fastcall TFarPanelInfo::GetSelectedCount()
{
  int Count = FPanelInfo->SelectedItemsNumber;

  if (Count == 1 &&
      (FPanelInfo->SelectedItems[0].Flags & PPIF_SELECTED) == 0)
  {
    Count = 0;
  }
  
  return Count;
}
//---------------------------------------------------------------------------
TList * __fastcall TFarPanelInfo::GetItems()
{
  if (!FItems)
  {
    FItems = new TObjectList();
    for (int Index = 0; Index < FPanelInfo->ItemsNumber; Index++)
    {
      FItems->Add(new TFarPanelItem(&FPanelInfo->PanelItems[Index]));
    }
  }
  return FItems;
}
//---------------------------------------------------------------------------
int __fastcall TFarPanelInfo::IndexOfFileName(const AnsiString FileName)
{
  TList * AItems = Items;
  TFarPanelItem * PanelItem;
  for (int Index = 0; Index < AItems->Count; Index++)
  {
    PanelItem = static_cast<TFarPanelItem *>(AItems->Items[Index]);
    if (PanelItem->FileName == FileName)
    {
      return Index;
    }
  }
  return -1;
}
//---------------------------------------------------------------------------
void __fastcall TFarPanelInfo::ApplySelection()
{
  FOwner->FarControl(FCTL_SETSELECTION, FPanelInfo);
}
//---------------------------------------------------------------------------
TFarPanelItem * __fastcall TFarPanelInfo::GetFocusedItem()
{
  int Index = FocusedIndex;
  if ((Index >= 0) && (Items->Count > 0))
  {
    assert(Index < Items->Count);
    return (TFarPanelItem *)Items->Items[Index];
  }
  else
  {
    return NULL;
  }
}
//---------------------------------------------------------------------------
void __fastcall TFarPanelInfo::SetFocusedItem(TFarPanelItem * value)
{
  int Index = Items->IndexOf(value);
  assert(Index >= 0);
  FocusedIndex = Index;
}
//---------------------------------------------------------------------------
int __fastcall TFarPanelInfo::GetFocusedIndex()
{
  return FPanelInfo->CurrentItem;
}
//---------------------------------------------------------------------------
void __fastcall TFarPanelInfo::SetFocusedIndex(int value)
{
  if (FocusedIndex != value)
  {
    assert(value >= 0 && value < FPanelInfo->ItemsNumber);
    FPanelInfo->CurrentItem = value;
    PanelRedrawInfo PanelInfo;
    PanelInfo.CurrentItem = FPanelInfo->CurrentItem;
    PanelInfo.TopPanelItem = FPanelInfo->TopPanelItem;
    FOwner->FarControl(FCTL_REDRAWPANEL, &PanelInfo);
  }
}
//---------------------------------------------------------------------------
//---------------------------------------------------------------------------
__fastcall TFarMenuItems::TFarMenuItems() :
  TStringList()
{
  FItemFocused = -1;
}
//---------------------------------------------------------------------------
void __fastcall TFarMenuItems::Clear()
{
  FItemFocused = -1;
  TStringList::Clear();
}
//---------------------------------------------------------------------------
void __fastcall TFarMenuItems::Delete(int Index)
{
  if (Index == FItemFocused)
  {
    FItemFocused = -1;
  }
  TStringList::Delete(Index);
}
//---------------------------------------------------------------------------
void __fastcall TFarMenuItems::PutObject(int Index, TObject * AObject)
{
  TStringList::PutObject(Index, AObject);
  bool Focused = (reinterpret_cast<int>(AObject) & MIF_SEPARATOR) != 0;
  if ((Index == ItemFocused) && !Focused)
  {
    FItemFocused = -1;
  }
  if (Focused)
  {
    if (ItemFocused >= 0)
    {
      SetFlag(ItemFocused, MIF_SELECTED, false);
    }
    FItemFocused = Index;
  }
}
//---------------------------------------------------------------------------
void __fastcall TFarMenuItems::AddSeparator()
{
  Add("");
  SetFlag(Count-1, MIF_SEPARATOR, true);
}
//---------------------------------------------------------------------------
void __fastcall TFarMenuItems::SetItemFocused(int value)
{
  if (ItemFocused != value)
  {
    if (ItemFocused >= 0)
    {
      SetFlag(ItemFocused, MIF_SELECTED, false);
    }
    FItemFocused = value;
    SetFlag(ItemFocused, MIF_SELECTED, true);
  }
}
//---------------------------------------------------------------------------
void __fastcall TFarMenuItems::SetFlag(int Index, int Flag, bool Value)
{
  if (GetFlag(Index, Flag) != Value)
  {
    int F = int(Objects[Index]);
    if (Value)
    {
      F |= Flag;
    }
    else
    {
      F &= ~Flag;
    }
    Objects[Index] = (TObject *)F;
  }
}
//---------------------------------------------------------------------------
bool __fastcall TFarMenuItems::GetFlag(int Index, int Flag)
{
  return int(Objects[Index]) & Flag;
}

